from enum import Enum, auto

from sampo.schemas.graph import GraphNode

StageType = tuple[GraphNode, dict[str, GraphNode]]


class SyntheticGraphType(Enum):
    """
    Describe available types of synthetic graph

    * PARALLEL - work graph dominated by parallel works
    * SEQUENTIAL - work graph dominated by sequential works
    * GENERAL - work graph, including sequential and parallel works, it is similar to the real work graphs
    """
    GENERAL = auto()
    PARALLEL = auto()
    SEQUENTIAL = auto()



from sampo.schemas.graph import GraphNode
from collections import deque


def count_ancestors(first_ancestors: list[GraphNode], root: GraphNode) -> int:
    """
    Counts the number of ancestors of the whole graph.

    :param first_ancestors: First ancestors of ancestors which must be counted.
    :param root: The root node of the graph.
    :return:
    """
    q = deque(first_ancestors)
    count = len(first_ancestors)
    used = set()
    used.add(root)
    while q:
        node = q.pop()
        for parent in node.parents:
            if parent in used:
                continue
            used.add(parent)
            q.appendleft(parent)
            count += 1

    return count



from enum import Enum, auto

from sampo.schemas.graph import GraphNode

StageType = tuple[GraphNode, dict[str, GraphNode]]


class SyntheticGraphType(Enum):
    """
    Describe types of synthetic graph:
    - Parallel - work graph in which has work is parallel to each other
    - Sequential - work graph in which the work is sequential to each other
    - General - work graph that closely resemble real work graphs in structure.
    """
    GENERAL = auto()
    PARALLEL = auto()
    SEQUENTIAL = auto()



from sampo.schemas.graph import GraphNode, EdgeType, WorkGraph
from sampo.schemas.works import WorkUnit
from sampo.structurator.prepare_wg_copy import prepare_work_graph_copy, new_start_finish


def insert_work_unit(original_wg: WorkGraph, inserted_wu: WorkUnit,
                     parents_edges: list[GraphNode] or list[tuple[GraphNode, float, EdgeType]],
                     children_edges: list[GraphNode] or list[tuple[GraphNode, float, EdgeType]],
                     change_id: bool = True) -> WorkGraph:
    """
    Inserts new node in the WorkGraph, based on given WorkUnit
    :param original_wg: WorkGraph into which we insert new node
    :param inserted_wu: WorkUnit on the basis of which we create new GraphNode
    :param parents_edges: nodes which are supposed to be the parents of new GraphNode
    :param children_edges: nodes which are supposed to be the children of new GraphNode
    :param change_id: do ids in the new graph need to be changed
    :return: new WorkGraph with inserted new node
    """
    reduced_parent_edges = _reduce_to_tuple_type(parents_edges)
    reduced_children_edges = _reduce_to_tuple_type(children_edges)

    copied_nodes, original_old_to_new_ids = prepare_work_graph_copy(original_wg, change_id=change_id)

    new_parents_edges = _new_edges(copied_nodes, original_old_to_new_ids, reduced_parent_edges)
    new_children_edges = _new_edges(copied_nodes, original_old_to_new_ids, reduced_children_edges)

    new_node = GraphNode(inserted_wu, new_parents_edges)
    for child, lag, edge in new_children_edges:
        child.add_parents([(new_node, lag, edge)])

    new_start, new_finish = new_start_finish(original_wg, copied_nodes, original_old_to_new_ids)

    return WorkGraph(new_start, new_finish)


def _new_edges(copied_nodes: dict[str, GraphNode], original_old_to_new_ids: dict[str, str],
               edges: list[tuple[GraphNode, float, EdgeType]]) \
        -> list[tuple[GraphNode, float, EdgeType]]:
    return [(copied_nodes[original_old_to_new_ids[parent.id]], lag, edge_type)
            for parent, lag, edge_type in edges]


def _reduce_to_tuple_type(edges: list[GraphNode] or list[tuple[GraphNode, float, EdgeType]]) \
        -> list[tuple[GraphNode, float, EdgeType]]:
    if isinstance(edges[0], GraphNode):
        return [(edge, -1, EdgeType.FinishStart) for edge in edges]
    else:
        return edges





from typing import Callable
from uuid import uuid4

from sampo.schemas.graph import GraphNode
from sampo.schemas.resources import Worker
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator
from sampo.schemas.works import WorkUnit


def calculate_working_time_cascade(node: GraphNode, appointed_workers: list[Worker],
                                   work_estimator: WorkTimeEstimator) -> Time:
    """
    Calculate the working time of the appointed workers at a current job for prioritization.
    O(1) - at worst case |inseparable_edges|

    :param appointed_worker:
    :param work_estimator:
    :param node: the target node
    :return: working time
    """
    if node.is_inseparable_son():
        # time of employment of resources and task execution is calculated only for the first job
        # in the chain of connected inextricably
        return Time(0)

    # calculation of the time for all work_units inextricably linked to the given
    common_time = Time(0)
    for dep_node in node.get_inseparable_chain_with_self():
        common_time += work_estimator.estimate_time(dep_node.work_unit, appointed_workers)
    return common_time


def calculate_working_time(work_unit: WorkUnit, appointed_worker: list[Worker],
                           work_estimator: WorkTimeEstimator = DefaultWorkEstimator) -> Time:
    """
    Calculate the working time of the appointed workers at a current job for final schedule

    :return: working time
    """
    return work_estimator.estimate_time(work_unit, appointed_worker)  # working time


PRIORITY_DELTA = 1


def work_priority(node: GraphNode,
                  comp_cost: Callable[[GraphNode, list[Worker], WorkTimeEstimator], Time],
                  work_estimator: WorkTimeEstimator) -> float:
    """
    Calculate the average time to complete the work when assigning the minimum and maximum number of employees
    for the correct calculations of rank in prioritization
    O(sum_of_max_counts_of_workers) of current work

    :param node: the target node
    :param work_estimator:
    :param comp_cost: function for calculating working time (calculate_working_time)
    :return: average working time
    """

    work_unit = node.work_unit

    passed_workers_min = [Worker(str(uuid4()), req.kind, req.min_count)
                          for req in work_unit.worker_reqs]

    passed_workers_max = [Worker(str(uuid4()), req.kind, req.max_count)
                          for req in work_unit.worker_reqs]

    res = (comp_cost(node, passed_workers_min, work_estimator) +
           comp_cost(node, passed_workers_max, work_estimator)) / 2

    return res + PRIORITY_DELTA



import numpy as np


# Here is the fastest stack. No checks will be performed in runtime.
class Stack(object):
    def __init__(self, max_length, dtype=np.float32):
        self.data = np.zeros(max_length, dtype=dtype)
        self.length = 0

    def push(self, x):
        self.data[self.length] = x
        self.length += 1

    def pop(self):
        self.length -= 1
        return self.data[self.length]

    def length(self):
        return self.data.size

    def swap(self, ind1: int, ind2: int):
        self.data[ind1], self.data[ind2] = self.data[ind2], self.data[ind1]

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, idx, item):
        self.data[idx] = item



from typing import Optional

from sampo.scheduler.timeline.base import Timeline
from sampo.scheduler.timeline.material_timeline import SupplyTimeline
from sampo.scheduler.timeline.zone_timeline import ZoneTimeline
from sampo.schemas.contractor import Contractor, WorkerContractorPool
from sampo.schemas.graph import GraphNode
from sampo.schemas.landscape import LandscapeConfiguration
from sampo.schemas.resources import Worker
from sampo.schemas.schedule_spec import WorkSpec
from sampo.schemas.scheduled_work import ScheduledWork
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator


class JustInTimeTimeline(Timeline):
    """
    Timeline that stored the time of resources release.
    For each contractor and worker type store a descending list of pairs of time and
    number of available workers of this type of this contractor.
    """

    def __init__(self, worker_pool: WorkerContractorPool, landscape: LandscapeConfiguration):
        self._timeline = {}
        # stacks of time(Time) and count[int]
        for worker_type, worker_offers in worker_pool.items():
            for worker_offer in worker_offers.values():
                self._timeline[worker_offer.get_agent_id()] = [(Time(0), worker_offer.count)]

        self._material_timeline = SupplyTimeline(landscape)
        self.zone_timeline = ZoneTimeline(landscape.zone_config)

    def find_min_start_time_with_additional(self, node: GraphNode,
                                            worker_team: list[Worker],
                                            node2swork: dict[GraphNode, ScheduledWork],
                                            spec: WorkSpec,
                                            assigned_start_time: Time | None = None,
                                            assigned_parent_time: Time = Time(0),
                                            work_estimator: WorkTimeEstimator = DefaultWorkEstimator()) \
            -> tuple[Time, Time, dict[GraphNode, tuple[Time, Time]]]:
        """
        Define the nearest possible start time for the current job. It is equal the max value from:
        1. end time of all parent tasks,
        2. time previous job off all needed workers to complete the current task.

        :param assigned_parent_time: minimum start time
        :param assigned_start_time:
        :param node: the GraphNode whose minimum time we are trying to find
        :param worker_team: the worker team under testing
        :param node2swork: dictionary, that match GraphNode to ScheduleWork respectively
        :param spec: given work specification
        :param work_estimator: function that calculates execution time of the GraphNode
        :return: start time, end time, None(exec_times not needed in this timeline)
        """
        # if current job is the first
        if not node2swork:
            max_material_time = self._material_timeline.find_min_material_time(node.id, assigned_parent_time,
                                                                               node.work_unit.need_materials(),
                                                                               node.work_unit.workground_size)

            max_zone_time = self.zone_timeline.find_min_start_time(node.work_unit.zone_reqs, max_material_time, Time(0))

            return max_zone_time, max_zone_time, None
        # define the max end time of all parent tasks
        max_parent_time = max(node.min_start_time(node2swork), assigned_parent_time)
        # define the max agents time when all needed workers are off from previous tasks
        max_agent_time = Time(0)

        if spec.is_independent:
            # grab from the end
            for worker in worker_team:
                offer_stack = self._timeline[worker.get_agent_id()]
                max_agent_time = max(max_agent_time, offer_stack[0][0])
        else:
            # grab from whole sequence
            # for each resource type
            for worker in worker_team:
                needed_count = worker.count
                offer_stack = self._timeline[worker.get_agent_id()]
                # traverse list while not enough resources and grab it
                ind = len(offer_stack) - 1
                while needed_count > 0:
                    offer_time, offer_count = offer_stack[ind]
                    max_agent_time = max(max_agent_time, offer_time)

                    if needed_count < offer_count:
                        offer_count = needed_count
                    needed_count -= offer_count
                    ind -= 1

        c_st = max(max_agent_time, max_parent_time)

        new_finish_time = c_st
        for dep_node in node.get_inseparable_chain_with_self():
            # set start time as finish time of original work
            # set finish time as finish time + working time of current node with identical resources
            # (the same as in original work)
            # set the same workers on it
            # TODO Decide where this should be
            dep_parent_time = dep_node.min_start_time(node2swork)

            dep_st = max(new_finish_time, dep_parent_time)
            working_time = work_estimator.estimate_time(dep_node.work_unit, worker_team)
            new_finish_time = dep_st + working_time

        exec_time = new_finish_time - c_st

        max_material_time = self._material_timeline.find_min_material_time(node.id, c_st,
                                                                           node.work_unit.need_materials(),
                                                                           node.work_unit.workground_size)

        max_zone_time = self.zone_timeline.find_min_start_time(node.work_unit.zone_reqs, c_st, exec_time)

        c_st = max(c_st, max_material_time, max_zone_time)

        c_ft = c_st + exec_time
        return c_st, c_ft, None

    def can_schedule_at_the_moment(self,
                                   node: GraphNode,
                                   worker_team: list[Worker],
                                   spec: WorkSpec,
                                   node2swork: dict[GraphNode, ScheduledWork],
                                   start_time: Time,
                                   exec_time: Time) -> bool:
        if spec.is_independent:
            # squash all the timeline to the last point
            for worker in worker_team:
                worker_timeline = self._timeline[(worker.contractor_id, worker.name)]
                last_cpkt_time, _ = worker_timeline[0]
                if last_cpkt_time >= start_time:
                    return False
            return True
        else:
            # checking edges
            for dep_node in node.get_inseparable_chain_with_self():
                for p in dep_node.parents:
                    if p != dep_node.inseparable_parent:
                        swork = node2swork.get(p, None)
                        if swork is None or swork.finish_time >= start_time:
                            return False

            max_agent_time = Time(0)
            for worker in worker_team:
                needed_count = worker.count
                offer_stack = self._timeline[worker.get_agent_id()]
                # traverse list while not enough resources and grab it
                ind = len(offer_stack) - 1
                while needed_count > 0:
                    offer_time, offer_count = offer_stack[ind]
                    max_agent_time = max(max_agent_time, offer_time)

                    if needed_count < offer_count:
                        offer_count = needed_count
                    needed_count -= offer_count
                    ind -= 1

            if not max_agent_time <= start_time:
                return False

            if not self._material_timeline.can_schedule_at_the_moment(node.id, start_time,
                                                                      node.work_unit.need_materials(),
                                                                      node.work_unit.workground_size):
                return False
            if not self.zone_timeline.can_schedule_at_the_moment(node.work_unit.zone_reqs, start_time, exec_time):
                return False

            return True

    def update_timeline(self,
                        finish_time: Time,
                        exec_time: Time,
                        node: GraphNode,
                        worker_team: list[Worker],
                        spec: WorkSpec):
        """
        Adds given `worker_team` to the timeline at the moment `finish`

        :param finish_time:
        :param exec_time:
        :param node:
        :param worker_team:
        :param spec: work specification
        :return:
        """
        if spec.is_independent:
            # squash all the timeline to the last point
            for worker in worker_team:
                worker_timeline = self._timeline[(worker.contractor_id, worker.name)]
                count_workers = sum([count for _, count in worker_timeline])
                worker_timeline.clear()
                worker_timeline.append((finish_time + 1, count_workers))
        else:
            # For each worker type consume the nearest available needed worker amount
            # and re-add it to the time when current work should be finished.
            # Addition performed as step in bubble-sort algorithm.
            for worker in worker_team:
                needed_count = worker.count
                worker_timeline = self._timeline[(worker.contractor_id, worker.name)]
                # Consume needed workers
                while needed_count > 0:
                    next_time, next_count = worker_timeline.pop()
                    if next_count > needed_count or len(worker_timeline) == 0:
                        worker_timeline.append((next_time, next_count - needed_count))
                        break
                    needed_count -= next_count

                # Add to the right place
                # worker_timeline.append((finish + 1, worker.count))
                # worker_timeline.sort(reverse=True)
                worker_timeline.append((finish_time + 1, worker.count))
                ind = len(worker_timeline) - 1
                while ind > 0 and worker_timeline[ind][0] > worker_timeline[ind - 1][0]:
                    worker_timeline[ind], worker_timeline[ind - 1] = worker_timeline[ind - 1], worker_timeline[ind]
                    ind -= 1

    def schedule(self,
                 node: GraphNode,
                 node2swork: dict[GraphNode, ScheduledWork],
                 workers: list[Worker],
                 contractor: Contractor,
                 spec: WorkSpec,
                 assigned_start_time: Optional[Time] = None,
                 assigned_time: Optional[Time] = None,
                 assigned_parent_time: Time = Time(0),
                 work_estimator: WorkTimeEstimator = DefaultWorkEstimator()):
        inseparable_chain = node.get_inseparable_chain_with_self()

        start_time = assigned_start_time if assigned_start_time is not None \
            else self.find_min_start_time(node, workers,
                                          node2swork,
                                          spec,
                                          assigned_parent_time,
                                          work_estimator)

        if assigned_time is not None:
            exec_times = {n: (Time(0), assigned_time // len(inseparable_chain))
                          for n in inseparable_chain}
            return self._schedule_with_inseparables(node, node2swork, workers, contractor, spec,
                                                    inseparable_chain, start_time, exec_times, work_estimator)
        else:
            return self._schedule_with_inseparables(node, node2swork, workers, contractor, spec,
                                                    inseparable_chain, start_time, {}, work_estimator)

    def _schedule_with_inseparables(self,
                                    node: GraphNode,
                                    node2swork: dict[GraphNode, ScheduledWork],
                                    workers: list[Worker],
                                    contractor: Contractor,
                                    spec: WorkSpec,
                                    inseparable_chain: list[GraphNode],
                                    start_time: Time,
                                    exec_times: dict[GraphNode, tuple[Time, Time]],
                                    work_estimator: WorkTimeEstimator = DefaultWorkEstimator()):
        """
        Makes ScheduledWork object from `GraphNode` and worker list, assigned `start_end_time`
        and adds it to given `node2swork`. Also does the same for all inseparable nodes starts from this one

        :param node:
        :param node2swork:
        :param workers:
        :param contractor:
        :param spec:
        :param inseparable_chain:
        :param start_time:
        :param exec_times:
        :param work_estimator:
        :return:
        """

        c_ft = start_time
        for dep_node in inseparable_chain:
            # set start time as finish time of original work
            # set finish time as finish time + working time of current node with identical resources
            # (the same as in original work)
            # set the same workers on it
            # TODO Decide where this should be
            max_parent_time = dep_node.min_start_time(node2swork)

            if dep_node.is_inseparable_son():
                assert max_parent_time >= node2swork[dep_node.inseparable_parent].finish_time

            if dep_node in exec_times:
                lag, working_time = exec_times[dep_node]
            else:
                lag, working_time = 0, work_estimator.estimate_time(node.work_unit, workers)
            c_st = max(c_ft + lag, max_parent_time)
            new_finish_time = c_st + working_time

            deliveries, _, new_finish_time = self._material_timeline.deliver_materials(dep_node.id, c_st,
                                                                                       new_finish_time,
                                                                                       dep_node.work_unit.need_materials(),
                                                                                       dep_node.work_unit.workground_size)

            node2swork[dep_node] = ScheduledWork(work_unit=dep_node.work_unit,
                                                 start_end_time=(c_st, new_finish_time),
                                                 workers=workers,
                                                 contractor=contractor,
                                                 materials=deliveries)
            # change finish time for using workers
            c_ft = new_finish_time

        zones = [zone_req.to_zone() for zone_req in node.work_unit.zone_reqs]
        self.update_timeline(c_ft, c_ft - start_time, node, workers, spec)
        node2swork[node].zones_pre = self.zone_timeline.update_timeline(len(node2swork), zones, start_time,
                                                                        c_ft - start_time)
        return c_ft



from typing import TypeVar, Generic

from sortedcontainers import SortedList

from sampo.schemas.time import Time
from sampo.schemas.types import EventType

T = TypeVar('T')


class GeneralTimeline(Generic[T]):
    """
    The representation of general-purpose timeline that supports some general subset of functions
    """
    def __init__(self):
        # ScheduleEvent = time, idx, object
        def event_cmp(event: Time | tuple[EventType, Time, int, T]) -> tuple[Time, int, int]:
            if isinstance(event, tuple):
                return event[1], event[2], event[0].priority

            if isinstance(event, Time):
                # instances of Time must be greater than almost all ScheduleEvents with same time point
                return event, Time.inf().value, 2

            raise ValueError(f'Incorrect type of value: {type(event)}')

        self._timeline = SortedList(iterable=((EventType.INITIAL, Time(0), -1, None),), key=event_cmp)
        self._next_idx = 0

    def update_timeline(self, start_time: Time, exec_time: Time, obj: T):
        self._timeline.add((EventType.START, start_time, self._next_idx, obj))
        self._timeline.add((EventType.END, start_time + exec_time, self._next_idx, obj))
        self._next_idx += 1

    def __getitem__(self, index) -> Time:
        """
        Returns the time of checkpoint on `index`
        """
        return self._timeline[index][1]

    def __len__(self):
        return len(self._timeline)



from abc import ABC, abstractmethod

import pandas as pd

from sampo.pipeline.lag_optimization import LagOptimizationStrategy
from sampo.scheduler.base import Scheduler
from sampo.scheduler.utils.local_optimization import OrderLocalOptimizer, ScheduleLocalOptimizer
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.landscape import LandscapeConfiguration
from sampo.schemas.project import ScheduledProject
from sampo.schemas.schedule import Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.utilities.task_name import NameMapper


class InputPipeline(ABC):
    """
    Base class to build different pipeline, that help to use the framework
    """

    @abstractmethod
    def wg(self, wg: WorkGraph | pd.DataFrame | str,
           full_connection: bool = False,
           change_base_on_history: bool = False) -> 'InputPipeline':
        ...

    @abstractmethod
    def contractors(self, contractors: list[Contractor] | pd.DataFrame | str) -> 'InputPipeline':
        ...

    @abstractmethod
    def name_mapper(self, name_mapper: NameMapper) -> 'InputPipeline':
        ...

    @abstractmethod
    def history(self, history: pd.DataFrame | str) -> 'InputPipeline':
        ...

    @abstractmethod
    def landscape(self, landscape_config: LandscapeConfiguration) -> 'InputPipeline':
        ...

    @abstractmethod
    def spec(self, spec: ScheduleSpec) -> 'InputPipeline':
        ...

    @abstractmethod
    def time_shift(self, time: Time) -> 'InputPipeline':
        ...

    @abstractmethod
    def lag_optimize(self, lag_optimize: LagOptimizationStrategy) -> 'InputPipeline':
        """
        Mandatory argument. Shows should graph be lag-optimized or not.
        If not defined, pipeline should search the best variant of this argument in result.

        :param lag_optimize:
        :return: the pipeline object
        """
        ...

    @abstractmethod
    def work_estimator(self, work_estimator: WorkTimeEstimator) -> 'InputPipeline':
        ...

    @abstractmethod
    def node_order(self, node_order: list[GraphNode]) -> 'InputPipeline':
        ...

    @abstractmethod
    def optimize_local(self, optimizer: OrderLocalOptimizer, area: range) -> 'InputPipeline':
        ...

    @abstractmethod
    def schedule(self, scheduler: Scheduler) -> 'SchedulePipeline':
        ...


class SchedulePipeline(ABC):
    """
    The part of pipeline, that manipulates with the whole entire schedule.
    """

    @abstractmethod
    def optimize_local(self, optimizer: ScheduleLocalOptimizer, area: range) -> 'SchedulePipeline':
        ...

    @abstractmethod
    def finish(self) -> ScheduledProject:
        ...



from sampo.userinput.parser.general_build import Graph
from typing import Iterable


def detect_cycles(input_graph: Iterable[object]) -> list | None:
    """
    Return list of cycles.

    :param input_graph: graph in abstract iterative form
    :return:
    """

    graph = Graph()

    for edge in input_graph:
        edge_info = tuple(edge)
        graph.add_edge(edge_info[0], edge_info[1])
    cycles = graph.eliminate_cycles(False)

    return cycles
